@@ -1,39 +1,46 @@
 """
 This module implements a friendly (well, friendlier) interface between the raw JSON
 responses from JIRA and the Resource/dict abstractions provided by this library. Users
-will construct a JIRA object as described below.
+will construct a JIRA object as described below. Full API documentation can be found
+at: https://jira-python.readthedocs.org/en/latest/
 """
-
 from __future__ import print_function
 
 from functools import wraps
+
 import imghdr
 import mimetypes
 
 import copy
 import os
-import urllib
 import re
 import sys
 import string
 import tempfile
 import logging
 import requests
+import json
+
+from six import string_types
+from six.moves.html_parser import HTMLParser
+from six import print_ as print
+
+from requests_oauthlib import OAuth1
+from oauthlib.oauth1 import SIGNATURE_RSA
+
+from jira.exceptions import raise_on_error
+# JIRA specific resources
+from jira.resources import Resource, Issue, Comment, Project, Attachment, Component, Dashboard, Filter, Votes, Watchers, Worklog, IssueLink, IssueLinkType, IssueType, Priority, Version, Role, Resolution, SecurityLevel, Status, User, CustomFieldOption, RemoteLink
+# GreenHopper specific resources
+from jira.resources import GreenHopperResource, Board, Sprint
+from jira.resilientsession import ResilientSession
+
 try:
     from random import SystemRandom
     random = SystemRandom()
 except ImportError:
     import random
-if sys.version_info[0] == 3:
-    import html.parser as HTMLParser
-else:
-    import HTMLParser
-from requests_oauthlib import OAuth1
-from oauthlib.oauth1 import SIGNATURE_RSA
-import json
-from jira.exceptions import raise_on_error
-from jira.resources import Resource, Issue, Comment, Project, Attachment, Component, Dashboard, Filter, Votes, Watchers, Worklog, IssueLink, IssueLinkType, IssueType, Priority, Version, Role, Resolution, SecurityLevel, Status, User, CustomFieldOption, RemoteLink
-from jira.resources import Board, Sprint
+
 if 'pydevd' not in sys.modules:
     try:
         import grequests
@@ -88,6 +95,7 @@ class JIRA(object):
         "rest_path": "api",
         "rest_api_version": "2",
         "verify": True,
+        "resilient": False,
         "headers": {
             'X-Atlassian-Token': 'nocheck',
             #'Cache-Control': 'no-cache',
@@ -96,6 +104,8 @@ class JIRA(object):
         }
     }
 
+    JIRA_BASE_URL = '{server}/rest/api/{rest_api_version}/{path}'
+
     def __init__(self, options=None, basic_auth=None, oauth=None, validate=None):
         """
         Construct a JIRA client instance.
@@ -116,6 +126,7 @@ class JIRA(object):
             * rest_path -- the root REST path to use. Defaults to ``api``, where the JIRA REST resources live.
             * rest_api_version -- the version of the REST resources under rest_path to use. Defaults to ``2``.
             * verify -- Verify SSL certs. Defaults to ``True``.
+            * resilient -- If it should just retry recoverable errors. Defaults to `False`.
         :param basic_auth: A tuple of username and password to use when establishing a session via HTTP BASIC
         authentication.
         :param oauth: A dict of properties for OAuth authentication. The following properties are required:
@@ -132,15 +143,9 @@ class JIRA(object):
 
         self._options = copy.copy(JIRA.DEFAULT_OPTIONS)
 
-        try:  # py3 variant
-            __import__("urllib.request", fromlist=[urllib])
-            self._options['proxies'] = urllib.request.getproxies()
-        except ImportError:  # py2 variant
-            self._options['proxies'] = urllib.getproxies()
-
         self._options.update(options)
 
-        # rip off trailing slash since all urls depend on that
+        # Rip off trailing slash since all urls depend on that
         if self._options['server'].endswith('/'):
             self._options['server'] = self._options['server'][:-1]
 
@@ -152,14 +157,16 @@ class JIRA(object):
             self._create_http_basic_session(*basic_auth)
         else:
             verify = self._options['verify']
-            self._session = requests.Session()
-            self._session.proxies = self._options['proxies']
+            if self._options['resilient']:
+                self._session = ResilientSession()
+            else:
+                self._session = requests.Session()
             self._session.verify = verify
             self._session.headers.update(self._options['headers'])
 
         if validate:
-            self.session()  # this will raise an Exception if you are not allowed to login. It's better to fail faster than later.
-        # we need version in order to know what API calls are available or not
+            self.session()  # This will raise an Exception if you are not allowed to login. It's better to fail faster than later.
+        # We need version in order to know what API calls are available or not
         self.version = tuple(self.server_info()['versionNumbers'])
 
 # Information about this client
@@ -237,7 +244,7 @@ class JIRA(object):
 
     def applicationlinks(self):
         """
-        list of application links
+        List of application links
         :return: json
         """
 
@@ -276,7 +283,7 @@ class JIRA(object):
             that a name is specified in one way or the other.
         :rtype: an Attachment Resource
         """
-        if isinstance(attachment, type('')):
+        if isinstance(attachment, string_types):
             attachment = open(attachment, "rb")
         # TODO: Support attaching multiple files at once?
         url = self._get_url('issue/' + str(issue) + '/attachments')
@@ -312,7 +319,7 @@ class JIRA(object):
     def create_component(self, name, project, description=None, leadUserName=None, assigneeType=None,
                          isAssigneeTypeValid=False):
         """
-        Create an issue component inside a project and return a Resource for it.
+        Create a component inside a project and return a Resource for it.
 
         :param name: name of the component
         :param project: key of the project to create the component in
@@ -666,7 +673,7 @@ class JIRA(object):
     @translate_resource_args
     def add_remote_link(self, issue, object, globalId=None, application=None, relationship=None):
         """
-        Create a remote link from an issue to an external application and returns a remote link Resource
+        Add a remote link from an issue to an external application and returns a remote link Resource
         for it. ``object`` should be a dict containing at least ``url`` to the linked external URL and
         ``title`` to display for the link inside JIRA.
 
@@ -835,9 +842,9 @@ class JIRA(object):
     def add_worklog(self, issue, timeSpent=None, adjustEstimate=None,
                     newEstimate=None, reduceBy=None, comment=None):
         """
-        Create a new worklog entry on an issue and return a Resource for it.
+        Add a new worklog entry on an issue and return a Resource for it.
 
-        :param issue: the issue to create the worklog on
+        :param issue: the issue to add the worklog to
         :param timeSpent: a worklog entry with this amount of time spent, e.g. "2d"
         :param adjustEstimate: (optional) allows the user to provide specific instructions to update the remaining\
         time estimate of the issue. The value can either be ``new``, ``leave``, ``manual`` or ``auto`` (default).
@@ -975,7 +982,7 @@ class JIRA(object):
             params['issueId'] = issueId
         return self._get_json('mypermissions', params=params)
 
-# PrioritiesK
+# Priorities
 
     def priorities(self):
         """Get a list of priority Resources from the server."""
@@ -1043,7 +1050,10 @@ class JIRA(object):
         :param boolean auto_confirm: whether to automatically confirm the temporary avatar by calling\
         :py:meth:`confirm_project_avatar` with the return value of this method.
         """
-        # TODO: autodetect size from passed-in file object?
+        size_from_file = os.path.getsize(filename)
+        if size != size_from_file:
+            size = size_from_file
+
         params = {
             'filename': filename,
             'size': size
@@ -1168,7 +1178,7 @@ class JIRA(object):
 
 # Search
 
-    def search_issues(self, jql_str, startAt=0, maxResults=50, fields=None, expand=None):
+    def search_issues(self, jql_str, startAt=0, maxResults=50, fields=None, expand=None, json_result=None):
         """
         Get a ResultList of issue Resources matching a JQL search string.
 
@@ -1198,6 +1208,8 @@ class JIRA(object):
             "fields": fields,
             "expand": expand
         }
+        if json_result:
+            return self._get_json('search', search_params)
 
         resource = self._get_json('search', search_params)
         issues = [Issue(self._options, self._session, raw_issue_json) for raw_issue_json in resource['issues']]
@@ -1290,7 +1302,7 @@ class JIRA(object):
         assignees, specify an issue key.
 
         :param username: a string to match usernames against
-        :param project: filter returned users by permission in this project (expected if a result will be used to\
+        :param project: filter returned users by permission in this project (expected if a result will be used to \
         create an issue)
         :param issueKey: filter returned users by this issue (expected if a result will be used to edit this issue)
         :param expand: extra information to fetch inside each resource
@@ -1346,7 +1358,10 @@ class JIRA(object):
         :param auto_confirm: whether to automatically confirm the temporary avatar by calling\
         :py:meth:`confirm_user_avatar` with the return value of this method.
         """
-        # TODO: autodetect size from passed-in file object?
+        size_from_file = os.path.getsize(filename)
+        if size != size_from_file:
+            size = size_from_file
+
         params = {
             'username': user,
             'filename': filename,
@@ -1541,7 +1556,14 @@ class JIRA(object):
     def session(self):
         """Get a dict of the current authenticated user's session information."""
         url = '{server}/rest/auth/1/session'.format(**self._options)
-        r = self._session.get(url)
+
+        if type(self._session.auth) is tuple:
+            authentication_data = {'username': self._session.auth[0], 'password': self._session.auth[1]}
+            r = self._session.post(url, data=json.dumps(authentication_data),
+                                   headers={'Content-type': 'application/json'})
+        else:
+            r = self._session.get(url)
+
         raise_on_error(r)
 
         user = User(self._options, self._session, json.loads(r.text))
@@ -1564,8 +1586,10 @@ class JIRA(object):
 # Utilities
     def _create_http_basic_session(self, username, password):
         verify = self._options['verify']
-        self._session = requests.Session()
-        self._session.proxies = self._options['proxies']
+        if self._options['resilient']:
+            self._session = ResilientSession()
+        else:
+            self._session = requests.Session()
         self._session.verify = verify
         self._session.auth = (username, password)
 
@@ -1578,8 +1602,10 @@ class JIRA(object):
             resource_owner_key=oauth['access_token'],
             resource_owner_secret=oauth['access_token_secret']
         )
-        self._session = requests.Session()
-        self._session.proxies = self._options['proxies']
+        if self._options['resilient']:
+            self._session = ResilientSession()
+        else:
+            self._session = requests.Session()
         self._session.verify = verify
         self._session.auth = oauth
 
@@ -1590,13 +1616,13 @@ class JIRA(object):
         r = self._session.put(url, params=params, headers={'content-type': 'application/json'}, data=json.dumps(data))
         raise_on_error(r)
 
-    def _get_url(self, path):
+    def _get_url(self, path, base=JIRA_BASE_URL):
         options = self._options
         options.update({'path': path})
-        return '{server}/rest/api/{rest_api_version}/{path}'.format(**options)
+        return base.format(**options)
 
-    def _get_json(self, path, params=None):
-        url = self._get_url(path)
+    def _get_json(self, path, base=JIRA_BASE_URL, params=None):
+        url = self._get_url(path, base)
         r = self._session.get(url, params=params, headers=self._options['headers'])
         raise_on_error(r)
 
@@ -1639,7 +1665,7 @@ class JIRA(object):
                       ". Specify the 'contentType' parameter explicitly.")
                 return None
 
-    def email_user(self, user, body, title="Jira Notification"):
+    def email_user(self, user, body, title="JIRA Notification"):
         """
         TBD:
         """
@@ -1667,7 +1693,7 @@ class JIRA(object):
 
     def rename_user(self, old_user, new_user):
         """
-        Rename a Jira user. Current implementation relies on third party plugin but in the future it may use embedded Jira functionality.
+        Rename a JIRA user. Current implementation relies on third party plugin but in the future it may use embedded JIRA functionality.
 
         :param old_user: string with username login
         :param new_user: string with username login
@@ -1756,9 +1782,9 @@ class JIRA(object):
         """
         Start jira re-indexing. Returns True if reindexing is in progress or not needed, or False.
 
-        If you call reindex() without any parameters it will perform a backfround reindex only if Jira thinks it should do it.
+        If you call reindex() without any parameters it will perform a backfround reindex only if JIRA thinks it should do it.
 
-        :param force: reindex even if Jira doesn'tt say this is needed, False by default.
+        :param force: reindex even if JIRA doesn'tt say this is needed, False by default.
         :param background: reindex inde background, slower but does not impact the users, defaults to True.
         """
         # /secure/admin/IndexAdmin.jspa
@@ -1772,7 +1798,7 @@ class JIRA(object):
 
         r = self._session.get(url, headers=self._options['headers'])
         if r.status_code == 503:
-            # logging.warning("Jira returned 503, this could mean that a full reindex is in progress.")
+            # logging.warning("JIRA returned 503, this could mean that a full reindex is in progress.")
             return 503
         # raise_on_error(r)
 
@@ -1780,7 +1806,7 @@ class JIRA(object):
             return True
 
         if r.content.find('All issues are being re-indexed'):
-            logging.warning("Jira re-indexing is already running.")
+            logging.warning("JIRA re-indexing is already running.")
             return True  # still reindexing is considered still a success
 
         if r.content.find('To perform the re-index now, please go to the') or force:
@@ -1793,9 +1819,9 @@ class JIRA(object):
                 return False
 
     def backup(self, filename='backup.zip'):
-        '''
+        """
         Will call jira export to backup as zipped xml. Returning with success does not mean that the backup process finished.
-        '''
+        """
         url = self._options['server'] + '/secure/admin/XmlBackup.jspa'
         payload = {'filename': filename}
         r = self._session.post(url, headers=self._options['headers'], data=payload)
@@ -1822,9 +1848,9 @@ class JIRA(object):
         return self._serverInfo['username']
 
     def delete_project(self, pid):
-        '''
+        """
         Project can be id, project key or project name. It will return False if it fails.
-        '''
+        """
         found = False
         try:
             if not str(int(pid)) == pid:
@@ -1849,12 +1875,12 @@ class JIRA(object):
             return r.status_code
 
     def create_project(self, key, name=None, assignee=None):
-        '''
+        """
         Key is mandatory and has to match JIRA project key requirements, usually only 2-10 uppercase characters.
         If name is not specified it will use the key value.
         If assignee is not specified it will use current user.
         The returned value should evaluate to False if it fails otherwise it will be the new project id.
-        '''
+        """
         if assignee is None:
             assignee = self.current_user()
         if name is None:
@@ -1930,9 +1956,10 @@ class JIRA(object):
             else:
                 return True
         return False
-    # experimental support for iDalko Grid, expect API to change as it's using private APIs currently
-    # https://support.idalko.com/browse/IGRID-1017
 
+# Experimental
+    # Experimental support for iDalko Grid, expect API to change as it's using private APIs currently
+    # https://support.idalko.com/browse/IGRID-1017
     def get_igrid(self, issueid, customfield, schemeid):
         url = self._options['server'] + '/rest/idalko-igrid/1.0/datagrid/data'
         if str(customfield).isdigit():
@@ -1958,196 +1985,192 @@ class JIRA(object):
 
 class GreenHopper(JIRA):
 
-    '''
+    """
     Define a class to hold functions for accessing GreenHopper resources.
     Extend the python-jira JIRA class.
-    '''
+    """
+
+    GREENHOPPER_BASE_URL = '{server}/rest/greenhopper/1.0/{path}'
 
     def __init__(self, options=None, basic_auth=None, oauth=None):
         self._rank = None
         JIRA.__init__(self, options, basic_auth, oauth)
 
-    def _gh_get_url(self, path):
-        ''' Use the given path for GH REST resources '''
-        options = self._options
-        options.update({'path': path})
-        return '{server}/rest/greenhopper/1.0/{path}'.format(**options)
+    def find(self, resource_format, ids=None):
+        """
+        Get a GreenHopperResource object for any addressable resource on the server.
 
-    def _gh_get_json(self, path, params=None):
-        ''' Return the GH data '''
-        url = self._gh_get_url(path)
-        r = self._session.get(url, params=params)
-        raise_on_error(r)
+        This method is a universal resource locator for any RESTful resource in GreenHopper. The
+        argument ``resource_format`` is a string of the form ``resource``, ``resource/{0}``,
+        ``resource/{0}/sub``, ``resource/{0}/sub/{1}``, etc. The format placeholders will be
+        populated from the ``ids`` argument if present. The existing authentication session
+        will be used.
 
-        r_json = json.loads(r.text)
-        return r_json
+        The return value is an untyped Resource object, which will not support specialized
+        :py:meth:`.Resource.update` or :py:meth:`.Resource.delete` behavior. Moreover, it will
+        not know to return an issue Resource if the client uses the resource issue path. For this
+        reason, it is intended to support resources that are not included in the standard
+        Atlassian REST API.
 
-    '''
-    Define the functions that interact with GreenHopper
-    '''
+        :param resource_format: the subpath to the resource string
+        :param ids: values to substitute in the ``resource_format`` string
+        :type ids: tuple or None
+        """
+        resource = GreenHopperResource(resource_format, self._options, self._session)
+        resource.find(ids)
+        return resource
 
+    """
+    Define the functions that interact with GreenHopper.
+    """
+
+    @translate_resource_args
     def boards(self):
-        '''
-        Return a list of all the boards
-        Example: rest/greenhopper/1.0/rapidviews/list
-        '''
-        r_json = self._gh_get_json('rapidviews/list')
-        boards = [Board(self._options, self._session, raw_res_json) for raw_res_json in r_json['views']]
+        """
+        Get a list of board GreenHopperResources.
+        """
+        r_json = self._get_json('rapidviews/list', base=self.GREENHOPPER_BASE_URL)
+
+        boards = [Board(self._options, self._session, raw_boards_json) for raw_boards_json in r_json['views']]
         return boards
 
+    @translate_resource_args
     def sprints(self, id):
-        '''
-        Return the Sprints that appear with the given board id
+        """
+        Get a list of sprint GreenHopperResources.
 
-        Example: rest/greenhopper/1.0/sprintquery/2
-        '''
-        # this fix is to handle the new API
+        :param id: the board to get sprints from
+        """
         r_json = {}
         try:
-            r_json = self._gh_get_json('sprintquery/%s' % id)
+            r_json = self._get_json('sprintquery/%s' % id, base=self.GREENHOPPER_BASE_URL)
         except:
-            r_json = self._gh_get_json('sprints/%s' % id)
+            r_json = self._get_json('sprints/%s' % id, base=self.GREENHOPPER_BASE_URL)
 
-        sprints = [Sprint(self._options, self._session, raw_res_json) for raw_res_json in r_json['sprints']]
+        sprints = [Sprint(self._options, self._session, raw_sprints_json) for raw_sprints_json in r_json['sprints']]
         return sprints
 
     def completed_issues(self, board_id, sprint_id):
-        '''
-        Return the completed issues for the given board id and sprint id
-        '''
+        """
+        Return the completed issues for ``board_id`` and ``sprint_id``.
+
+        :param board_id: the board retrieving issues from
+        :param sprint_id: the sprint retieving issues from
+        """
         # TODO need a better way to provide all the info from the sprintreport
         # incompletedIssues went to backlog but not it not completed
         # issueKeysAddedDuringSprint used to mark some with a * ?
         # puntedIssues are for scope change?
 
-        r_json = self._gh_get_json('rapid/charts/sprintreport?rapidViewId=%s&sprintId=%s' % (board_id, sprint_id))
-        issues = [Issue(self._options, self._session, raw_res_json) for raw_res_json in r_json['contents']['completedIssues']]
+        r_json = self._get_json('rapid/charts/sprintreport?rapidViewId=%s&sprintId=%s' % (board_id, sprint_id), base=self.GREENHOPPER_BASE_URL)
+        issues = [Issue(self._options, self._session, raw_issues_json) for raw_issues_json in r_json['contents']['completedIssues']]
         return issues
 
     def incompleted_issues(self, board_id, sprint_id):
-        '''
+        """
         Return the completed issues for the given board id and sprint id
-        '''
-        r_json = self._gh_get_json('rapid/charts/sprintreport?rapidViewId=%s&sprintId=%s' % (board_id, sprint_id))
-        issues = [Issue(self._options, self._session, raw_res_json) for raw_res_json in r_json['contents']['incompletedIssues']]
+        """
+        r_json = self._get_json('rapid/charts/sprintreport?rapidViewId=%s&sprintId=%s' % (board_id, sprint_id), self.GREENHOPPER_BASE_URL)
+        issues = [Issue(self._options, self._session, raw_issues_json) for raw_issues_json in r_json['contents']['incompletedIssues']]
         return issues
 
     def sprint_info(self, board_id, sprint_id):
-        '''
+        """
         Return the information about a sprint.
-        This uses the same method as completed issues
-        '''
-        r_json = self._gh_get_json('rapid/charts/sprintreport?rapidViewId=%s&sprintId=%s' % (board_id, sprint_id))
-        return r_json['sprint']
+
+        :param board_id: the board retrieving issues from
+        :param sprint_id: the sprint retieving issues from
+        """
+        return self._get_json('rapid/charts/sprintreport?rapidViewId=%s&sprintId=%s' % (board_id, sprint_id), base=self.GREENHOPPER_BASE_URL)['sprint']
 
     def create_board(self, name, project_ids, preset="scrum"):
-        '''
-        Create a new board for the given projects
-        Preset can be kanban, scrum or diy
-        '''
+        """
+        Create a new board for the ``project_ids``.
+
+        :param name: name of the board
+        :param project_ids: the projects to create the board in
+        :param preset: what preset to use for this board
+        :type preset: 'kanban', 'scrum', 'diy'
+        """
         payload = {}
         payload['name'] = name
         payload['projectIds'] = project_ids
         payload['preset'] = preset
-        url = self._gh_get_url('rapidview/create/presets')
-        r = self._session.post(url, data=json.dumps(payload))
+        url = self._get_url('rapidview/create/presets', base=self.GREENHOPPER_BASE_URL)
+        r = self._session.post(url, headers={'content-type': 'application/json'}, data=json.dumps(payload))
         raise_on_error(r)
-        # This isn't really a Board object, just a subset for the id
-        result = r.json
-        return result
+
+        raw_issue_json = json.loads(r.text)
+        return Board(self._options, self._session, raw=raw_issue_json)
 
     def create_sprint(self, name, board_id):
-        '''
-        Create a new sprint for the given board
+        """
+        Create a new sprint for the ``board_id``.
 
-        createSprintWithIssues in CreateSprintResource
-
-        Long rapidViewId;
-        Long sprintMarkerId;
-        List<String> issuesForSprint;
-        String name;
-        String startDate;
-        String endDate;
-
-        This is before the sprint is started:
-        http://localhost:8080/rest/greenhopper/1.0/backlog/markers/add
-        {"rapidViewId":4}
-        which returns
-        {"afterMarkerId":10,"name":"Our First Sprint 4","id":11}
-
-        When an sprint is started:
-        {"rapidViewId":4,"sprintMarkerId":11,"issuesForSprint":["SRC-1"],"name":"Our First Sprint 5","startDate":"22/Mar/13 8:51 PM","endDate":"05/Apr/13 8:51 PM"}
-
-        Responds:
-        {"id":5,"name":"Our First Sprint 5","closed":false}
-
-        And create/model comes before create, and expects issues to add
-        {"rapidViewId":2,"sprintMarkerId":0} fails.
-
-        When a sprint is in backlog it is just a marker
-
-        {"rapidViewId":2,"sprintMarkerId":0,"issuesForSprint":[],"name":"SPR 1","startDate":"26/Mar/13 11:36 AM","endDate":"26/Mar/13 12:36 PM"}
-
-        '''
+        :param name: name of the sprint
+        :param board_id: the board to add the sprint to
+        """
         payload = {}
         payload['name'] = name
         payload['rapidViewId'] = board_id
         payload['sprintMarkerId'] = 0
-        url = self._gh_get_url('sprint/create')
-        r = self._session.post(url, data=json.dumps(payload))
+        url = self._get_url('sprint/create', base=self.GREENHOPPER_BASE_URL)
+        r = self._session.post(url, headers={'content-type': 'application/json'}, data=json.dumps(payload))
         raise_on_error(r)
-        # This isn't really a Sprint object, just a subset for the id
-        result = r.json
-        return result
+
+        raw_issue_json = json.loads(r.text)
+        return Sprint(self._options, self._session, raw=raw_issue_json)
 
     def add_issues_to_sprint(self, sprint_id, issue_keys):
-        '''
-        Add the issues in the array of issue keys to the given started
-        but not completed sprint. Idempotent.
+        """
+        Add the issues in ``issue_keys`` to the ``sprint_id``. The sprint must
+        be started but not completed.
 
-        If a sprint was completed then have to also edit the issues' history
-        so that it was added to the sprint before it was completed,
-        preferably before it started. A completed sprint's issues also
-        all have a resolution set before the completion date.
+        If a sprint was completed, then have to also edit the history of the
+        issue so that it was added to the sprint before it was completed,
+        preferably before it started. A completed sprint's issues also all have
+        a resolution set before the completion date.
 
-        If a sprint was not started then have to edit the marker
-        and copy the rank of each issue too.
+        If a sprint was not started, then have to edit the marker and copy the
+        rank of each issue too.
 
-        /sprint/{sprintId}/issues/add
-        SprintIssuesResource.java
-        @Path("add")
-        public Response addIssueToSprint(@PathParam("sprintId") final Long sprintId, final IssuesKeysModel model)
-        {"issueKeys":["TS-3"]}   (paste this into textarea, not as a custom parameter)
-        When this is working in the rest browser must check in Chrome to see if put or post. Post gave 405 Method Not Allowed
-        '''
+        :param sprint_id: the sprint to add issues to
+        :param issue_keys: the issues to add to the sprint
+        """
         data = {}
         data['issueKeys'] = issue_keys
-        url = self._gh_get_url('sprint/%s/issues/add' % (sprint_id))
+        url = self._get_url('sprint/%s/issues/add' % (sprint_id), base=self.GREENHOPPER_BASE_URL)
         r = self._session.put(url, data=json.dumps(data))
         raise_on_error(r)
 
     def add_issues_to_epic(self, epic_id, issue_keys, ignore_epics=True):
+        """
+        Add the issues in ``issue_keys`` to the ``epic_id``.
+
+        :param epic_id: the epic to add issues to
+        :param issue_keys: the issues to add to the epic
+        :param ignore_epics: ignore any issues listed in ``issue_keys`` that are epics
+        """
         data = {}
         data['issueKeys'] = issue_keys
         data['ignoreEpics'] = ignore_epics
-        url = self._gh_get_url('epics/%s/add' % epic_id)
+        url = self._get_url('epics/%s/add' % epic_id, base=self.GREENHOPPER_BASE_URL)
         r = self._session.put(url, data=json.dumps(data), headers={'Content-Type': 'application/json'})
         raise_on_error(r)
 
     def rank(self, issue, next_issue):
-        '''
+        """
         Rank an issue before another using the default Ranking field, the one named 'Rank'.
 
         :param issue: issue key of the issue to be ranked before the second one.
         :param next_issue: issue key of the second issue.
-
-        '''
+        """
         # {"issueKeys":["ANERDS-102"],"rankBeforeKey":"ANERDS-94","rankAfterKey":"ANERDS-7","customFieldId":11431}
         if not self._rank:
             for field in self.fields():
                 if field['name'] == 'Rank' and field['schema']['custom'] == "com.pyxis.greenhopper.jira:gh-global-rank":
                     self._rank = field['schema']['customId']
         data = {"issueKeys": [issue], "rankBeforeKey": next_issue, "customFieldId": self._rank}
-        url = self._gh_get_url('rank')
+        url = self._get_url('rank', base=self.GREENHOPPER_BASE_URL)
         r = self._session.put(url, data=json.dumps(data), headers={'Content-Type': 'application/json'})
         raise_on_error(r)
