@@ -2,26 +2,19 @@
 This module implements the Resource classes that translate JSON from JIRA REST resources
 into usable objects.
 """
+from __future__ import print_function
 
 import re
 import sys
 import logging
 import random
 import pprint
-from jira.exceptions import raise_on_error, get_error_list
 import json
 
-if sys.version_info[0] == 3:
-    basestring = str
-    iteritems = dict.items
-else:
-    iteritems = dict.iteritems
+from six import iteritems, string_types, text_type
+from six import print_ as print
 
-if 'pydevd' not in sys.modules:
-    try:
-        import grequests
-    except ImportError:
-        pass  # that's an optional experimental feature, so there is no need to force the dependency
+from jira.exceptions import raise_on_error, get_error_list
 
 
 class Resource(object):
@@ -60,7 +53,7 @@ class Resource(object):
         self._options = options
         self._session = session
 
-        # explicitly define as None so we know when a resource has actually been loaded
+        # Explicitly define as None so we know when a resource has actually been loaded
         self.raw = None
         self.self = None
 
@@ -69,10 +62,10 @@ class Resource(object):
         if self.raw:
             for name in self._READABLE_IDS:
                 if name in self.raw:
-                    pretty_name = unicode(self.raw[name])
+                    pretty_name = text_type(self.raw[name])
                     # Include any child to support nested select fields.
                     if hasattr(self, 'child'):
-                        pretty_name += ' - ' + unicode(self.child)
+                        pretty_name += ' - ' + text_type(self.child)
                     return pretty_name
 
         # If all else fails, use repr to make sure we get something.
@@ -87,14 +80,14 @@ class Resource(object):
                     names.append(name + '=' + repr(self.raw[name]))
         if not names:
             return '<JIRA %s at %s>' % (self.__class__.__name__,
-                                        unicode(hex(id(self))))
+                                        text_type(hex(id(self))))
         return '<JIRA %s: %s>' % (self.__class__.__name__, ', '.join(names))
 
     def find(self, ids=None, headers=None, params=None):
         if ids is None:
             ids = ()
 
-        if isinstance(ids, basestring):
+        if isinstance(ids, string_types):
             ids = (ids,)
 
         if headers is None:
@@ -158,17 +151,19 @@ class Resource(object):
 
             if user:
                 logging.warning("Trying to add missing orphan user '%s' in order to complete the previous failed operation." % user)
-                # print pprint.pprint(self.__dict__)
                 jira.add_user(user, 'noreply@example.com', 10100, active=False)
-                    # if 'assignee' not in data['fields']:
-                    #    logging.warning("autofix: setting assignee to '%s' and retrying the update." % self._options['autofix'])
-                    #    data['fields']['assignee'] = {'name': self._options['autofix']}
+                # if 'assignee' not in data['fields']:
+                #    logging.warning("autofix: setting assignee to '%s' and retrying the update." % self._options['autofix'])
+                #    data['fields']['assignee'] = {'name': self._options['autofix']}
+            # EXPERIMENTAL --->
+            # import grequests
             # if async and 'grequests' in sys.modules:
             #   if not hasattr(self._session, '_async_jobs'):
             #        self._session._async_jobs = set()
             #    self._session._async_jobs.add(grequests.put(self.self, headers={'content-type': 'application/json'}, data=json.dumps(data)))
             # else:
             #    print "x", data
+            # <--- EXPERIMENTAL
             r = self._session.put(self.self, headers={'content-type': 'application/json'}, data=json.dumps(data))
         raise_on_error(r)
         self._load(self.self)
@@ -484,9 +479,9 @@ class Role(Resource):
         :param groups: a group or groups to add to the role
         :type groups: string, list or tuple
         """
-        if users is not None and isinstance(users, basestring):
+        if users is not None and isinstance(users, string_types):
             users = (users,)
-        if groups is not None and isinstance(groups, basestring):
+        if groups is not None and isinstance(groups, string_types):
             groups = (groups,)
 
         data = {
@@ -567,6 +562,18 @@ class Version(Resource):
 
         super(Version, self).delete(params)
 
+    def update(self, **args):
+        """
+        Update this project version from the server. It is prior used to archive
+        versions
+        """
+        data = {}
+        for field in args:
+            data[field] = args[field]
+
+        super(Version, self).update(**data)
+
+
 # GreenHopper
 
 
@@ -641,6 +648,7 @@ def dict2resource(raw, top=None, options=None, session=None):
     return top
 
 resource_class_map = {
+    # JIRA specific resources
     r'attachment/[^/]+$': Attachment,
     r'component/[^/]+$': Component,
     r'customFieldOption/[^/]+$': CustomFieldOption,
@@ -662,6 +670,9 @@ resource_class_map = {
     r'status/[^/]+$': Status,
     r'user\?username.+$': User,
     r'version/[^/]+$': Version,
+    # GreenHopper specific resources
+    r'sprints/[^/]+$': Sprint,
+    r'views/[^/]+$': Board,
 }
 
 
@@ -670,5 +681,5 @@ def cls_for_resource(resource_literal):
         if re.search(resource, resource_literal):
             return resource_class_map[resource]
     else:
-        # generic Resource without specialized update/delete behavior
+        # Generic Resource without specialized update/delete behavior
         return Resource
